==> hakk_builder-0.2.1.tar.gz: lib/src/basic/flow_control.dart <==
part of hakkit.hakk.builder.basic;


/// Declares a looping statement that accomodates an [initializer], 
/// a terminating [whileCondition], and an [incrementorFunc] expression.
/// 
/// Used in conjunction with a [NEXT] statement to complete a FOR/NEXT block
/// of code, the FOR statement provides similar semantics to the standard
/// for loops seen in many languages.
/// 
/// # incrementor must pass a function
/// Because of the way the underlying builder system works, the incrementor 
/// parameter [incrementorFunc] (the third parameter), must be passed in as a
/// function.  If you are not familiar with Dart, just follow the example below
/// for the syntax.
/// 
/// # Example
///     hakk('for_test');
///     Block.system(0, 1);
///     
///     // nested FOR loops
///     FOR(core0 << 0, core0 < 10, () => inc(core0));
///       output('in core 0');
///       FOR(core1 << 0, core1 < 5, () => inc(core1));
///         output('   in core 1');
///       NEXT(); 
///     NEXT();
///     
/// # Related
/// * [NEXT]
/// * [WHILE]
/// * [ENDWHILE]
/// 
/// @param initializer initializes the memory location to the value that begins
/// the for/next sequence.
/// @param whileCondition is tested at the beginning of each iteration through the
/// FOR/NEXT loop, if the test returns false then the loop terminates.
/// @param incrementorFunc is a function that calls a given expression at the end
/// of the loop, just before returning to the top again.
/// 
void FOR(initializer, whileCondition, incrementorFunc()){
  whileCondition = _wrapLiteral(whileCondition);
  
  if (initializer != null){
    throw 'FOR initializer should be an assignment (ex. core0 << 0)';
  }
  
  _forStack.add(incrementorFunc);
  
  _globalContext.add('_for_context_${_nextLabelID}_');
  WHILE(whileCondition);
}


/// Declares the end of a FOR/NEXT loop block.
/// 
/// # Example
/// See the example in [FOR] to see how NEXT is used.
/// 
/// # Related
/// * [FOR]
/// * [WHILE]
/// * [ENDWHILE]
void NEXT(){
  if (_forStack.isEmpty){
    throw 'Found NEXT() without a corresponding FOR(). Check to make sure your FOR/NEXT declarations are balanced.';
  }
  
  if (_globalContext.isEmpty){
    throw 'Bad compiler state.  This is probably a bug on our end.';
  }
  
  
  final gctx = _globalContext.removeLast();
  final incrementor = _forStack.removeLast();
  incrementor();
  ENDWHILE();
  _MMU.dalloc('expression${gctx}');
  if (_globalContext.isEmpty){
    _MMU.dalloc('expression');
  }
}

/// Declares a condition loop that continues while the given [expr] evaluates to
/// true.
/// 
/// All WHILE() statements just eventually end with a [ENDWHILE] statement.
/// 
/// ## Limitations
/// Complex looping scenarios are not well supported.  Use care when nesting WHILE
/// loops, as this will likely yield unexpected results.  You may need to use the
/// advanced API if you require more complex looping implementations, or [inline]
/// your own hakkML directly.
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///          
///     // loop that ends when core0 >= 10
///     WHILE(core0 < 10);
///       output(core0);
///       inc(core0);
///     ENDWHILE();
///     
///     // infinite loop.  Very common in a hakk to have a main loop that
///     // is continuous.
///     WHILE(true);
///       output('to infinity and beyond!');
///     ENDWHILE();
/// 
/// # Related
/// * [ENDWHILE]
/// * [FOR]
/// * [NEXT]
void WHILE(expr){
  expr = _wrapLiteral(expr);
  
  _whileStack.add(_nextLabelID);
  _HakkObject.addLine('lbl: __while_test_${_whileStack.last}__');
  _globalContext.add('_while_context_${_nextLabelID}_');
  IF(expr);
}

/// Defines the end of a [WHILE] block.  When the [WHILE] expression evaluates to
/// false, code immediately following ENDWHILE will execute.
/// 
/// # Related
/// * [WHILE]
/// * [FOR]
/// * [NEXT]
void ENDWHILE(){
  if (_whileStack.isEmpty){
    throw 'Found ENDWHILE() without a corresponding WHILE(). Check to make sure your WHILE/ENDWHILE declarations are balanced.';
  }
  _HakkObject.addLine('goto: __while_test_${_whileStack.last}__');
  ENDIF();
  final gctx = _globalContext.removeLast();
  _MMU.dalloc('expression${gctx}');
  if (_globalContext.isEmpty){
    _MMU.dalloc('expression');
  }
  _whileStack.removeLast();
}

/// Provides a conditional mechanism that executes code inside the IF/ENDIF block 
/// when the given [expr] evaluates to true.
/// 
/// Any IF statement must have a corresponding [ENDIF].
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///     
///     core0 << 10
///     IF(core0 < 100);
///       output('core0 is less than 100');
///     ENDIF();
///     
/// # Related
/// * [ENDIF]
/// 
/// @param expr is the conditional expression to test.
void IF(expr){
  expr = _wrapLiteral(expr);
  _ifStack.add(_nextLabelID);
 
  final i = _ifStack.length;
  final li = _nextLabelID;

  final m = _MMU.malloc('if_context_${_ifStack.last}');
  
  if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('je: [${expr.memLoc}, true, __iftrue_${_ifStack.last}__]');
  }else{
    _HakkObject.addLine('set: [${m}, ${expr}]');
    _HakkObject.addLine('je: [${m}, true, __iftrue_${_ifStack.last}__]');    
  }
  _HakkObject.addLine('goto: __endif_${_ifStack.last}__');
  _HakkObject.addLine('lbl: __iftrue_${_ifStack.last}__');
  
  _globalContext.add('_if_context_${_nextLabelID}_');
}

/// Defines the end of a [IF] block.
/// 
/// # Related
/// * [IF]
void ENDIF(){
  if (_ifStack.isEmpty){
    throw 'Found ENDIF() without a corresponding IF(). Check to make sure your IF/ENDIF declarations are balanced.';
  }
  _HakkObject.addLine('lbl: __endif_${_ifStack.last}__');
  _MMU.dalloc('if_context_${_ifStack.last}');
  _ifStack.removeLast();
  final gctx = _globalContext.removeLast();
  _MMU.dalloc('expression${gctx}');
  if (_globalContext.isEmpty){
    _MMU.dalloc('expression');
  }
}

/// Jumps code execution to the given [label].  A label can be a local [label] in
/// the current instruction block context, or it can be the name of another 
/// [Block.instruction], allowing transfer of control between instruction blocks.
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///     
///     GOTO('this_place');
///     output('This will never be seen.');
///     
///     label('this_place');
///     output('Hi there.');
///     
/// @param label is the name of the label or instruction block to jump to.
void GOTO(String label){
  _HakkObject.addLine('goto: ${label}');
}

/// Initiates a call to a [SUB] subroutine, which is eventually [RETURN]ed to the
/// instruction immediately following the GOSUB statement.
/// 
/// # parameter passing and value returns
/// The GOSUB/SUB/RETURN sequence also supports a rudimentary parameter passing
/// and value return.
/// 
/// ## Subroutines are NOT closures.
/// All values stored in memory blocks (such as [core0] or any memory block added
/// to a hakk) are essential "global".  This means that you must plan carefully
/// the use of these memory locations to prevent "clobbering", where one operation
/// overwrites a memory location that was not intended.
/// 
/// ## param/argument matching
/// It is important that when using GOSUB the number of parameter values passed, must
/// match the number of arguements expected to be received by the [SUB]routine. 
///  
/// # GOSUB Must be used in an expression.
/// [SUB]routines always return a value, even if one is not provided.  The default
/// return value is 0, if none is provided.  A GOSUB should always be assigned to some
/// memory location.  It will not execute correctly if called alone.
/// 
/// ## This works
///     core0 << GOSUB('foo');
/// 
/// ## This won't work
///     GOSUB('foo');
/// 
/// ## If you want to call a [SUB] and do care about the return result, use [CALL].
///     CALL('foo');
///  
/// # Limitations
/// While GOSUB returns are well tolerated in expressions, they cannot at this time
/// be used directly as parameters.
/// 
/// ## Don't do this
///     GOSUB('foo', [GOSUB('bar'), GOSUB('bar')]);
///     
/// ## Do this instead
///     core0 << GOSUB('bar');
///     GOSUB('foo', [core0, core0]);
/// 
/// # Example
///     hakk('GOSUB_example');
///     Block.system(0, 1);
///     
///     // GOSUB to a subroutine with no parameters
///     core0 << GOSUB('say_hi');
///     
///     // Pass 2 parameters to 'adder' and store the
///     // returned result in core0.
///     core0 << GOSUB('adder', [5, 5]);
///     // "10"
///     output(core0);
///     
///     SUB('say_hi');
///       output('hi');
///     RETURN();
///     
///     SUB('adder', [core3, core4]);
///       core5 << core3 + core4;
///     RETURN(core5);
///     
///     output('Hi there.');
///     
/// # Related
/// * [SUB]
/// * [RETURN]
/// * [CALL]
/// 
/// @param subroutineName is the name of the [SUB]routine to call.
/// @param params optionally passes parameters to the [SUB]routine.
_ExpressionNode GOSUB(String subroutineName, [List params]){
  final node = new _ExpressionNode.call(subroutineName, params);
  _expressionQueue.add(node);
  return node;
}

/// A version of [GOSUB] that can be used outside of an expression.
/// 
/// # Example
///     hakk('CALL_example');
///     Block.system(0, 1);
///     
///     // CALL to a subroutine, but don't worry about the result returned.
///     CALL('foo');
///     
///     // Using GOSUB, we would have to assign it to a memory location or use it
///     // in an expression.
///     core0 << GOSUB('foo');
///     
///     // core0 now equals 0, because that is the default return value none is 
///     // provided by the called subroutine.
///     
///     output('finished');
///     
///     SUB('foo');
///       output('hello world!');
///     RETURN();
/// 
/// # Related
/// * [SUB]
/// * [RETURN]
/// * [GOSUB]
/// 
/// @param subroutineName is the name of the [SUB]routine to call.
/// @param params optionally passes parameters to the [SUB]routine.
void CALL(String subroutineName, [List params]){
  final node = new _ExpressionNode.call(subroutineName, params);
  _expressionQueue.add(node);
  node.compile();
}

/// Declares a subroutine block that can take 0-n arguments, run some code,
/// and then return a result back to the [GOSUB] that called it.
/// 
/// # Limitations
/// A Subroutine is NOT a closure.  All values and references used within it are
/// global to the hakk runtime context.
/// 
/// Subroutines do not support recursion (calling itself) or nesting (a subroutine 
/// within another subroutine).  Attempts to use them in these ways will likely 
/// yield unexpected results.  
/// 
/// It is possible to emulate these behaviors yourself, by using the advanced API, or
/// by [inline]ing hakkML.  Proceed with caution.
/// 
/// # Example
/// See the [GOSUB] documentation for examples of how SUB is used.
/// 
/// # Related
/// * [GOSUB]
/// * [RETURN]
/// * [CALL]
/// 
/// @param subroutineName is the declared name of the subroutine that [GOSUB] will use
/// to call it.
/// @param arguments is a list of memory locations that will store the values passed
/// by the [GOSUB] call.  Make sure the number of arguments matches the number of
/// parameters!
void SUB(String subroutineName, [List<_HakkObject> arguments]){
  if (_subroutineContext != null){
    throw 'Nested subroutines are not allowed.';
  }
  _subroutineContext = '${subroutineName}_protect';
  GOTO(_subroutineContext);
  label(subroutineName);
  if (arguments != null){
    arguments.reversed.forEach((a){
      if (a is! _HakkObject || (a is _HakkObject && a.isLiteral)){
        throw 'Argument must be a memory reference.';
      }
      a << pop();
    });
  }
}

/// Returns control from a [SUB]routine back to the [GOSUB] that called it.
/// 
/// RETURN may pass an optiona expression back to the GOSUB.
///
/// # Example
/// See the [GOSUB] documentation for examples of how RETURN is used.
///
/// # Related
/// * [SUB]
/// * [GOSUB]
void RETURN([expr]){
  if (_subroutineContext == null){
    throw 'Cannot have a subroutine [ret]urn without first declaring a subroutine().';
  }
  if (expr == null){
    _HakkObject.addLine('push: 0');
  }else{
    expr = _wrapLiteral(expr);
    push(expr);
  }
  
  _HakkObject.addLine('ret:');
  label(_subroutineContext);
  _subroutineContext = null;
}


/// Defines a label in code that can be used by [GOTO] to jump to a different
/// location in code.
/// 
/// # Example
/// See the [GOTO] documenation for examples of how label can be used.
/// 
/// # Related
/// * [GOTO]
/// 
/// @param label is the name which will be used by a [GOTO] statement to jump to this
/// location in the hakk.
void label(String label){
  _HakkObject.addLine('lbl: ${label}');
}


/// Unconditionally terminates processing of hakk code in the hakk.it virtual
/// machine.
/// 
/// This instruction is generally used in testing only.  It probably wouldn't help
/// your chances to win in the arenas if your hakk stops running!
/// 
/// @param expr is an expression that should result in a termination code for the
/// hakk.  At this time, this information is not exposed to the player, and can only
/// be inspected by internal tooling.
void EXIT(expr){
  expr = _wrapLiteral(expr);
  if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('exit: ${expr.memLoc}');
  }else{
    _HakkObject.addLine('exit: ${expr}');    
  }
}

==> hakk_builder-0.2.2.tar.gz: lib/src/basic/flow_control.dart <==
part of hakkit.hakk.builder.basic;


/// Declares a looping statement that accomodates an [initializer], 
/// a terminating [whileCondition], and an [incrementorFunc] expression.
/// 
/// Used in conjunction with a [NEXT] statement to complete a FOR/NEXT block
/// of code, the FOR statement provides similar semantics to the standard
/// for loops seen in many languages.
/// 
/// # incrementor must pass a function
/// Because of the way the underlying builder system works, the incrementor 
/// parameter [incrementorFunc] (the third parameter), must be passed in as a
/// function.  If you are not familiar with Dart, just follow the example below
/// for the syntax.
/// 
/// # Example
///     hakk('for_test');
///     Block.system(0, 1);
///     
///     // nested FOR loops
///     FOR(core0 << 0, core0 < 10, () => inc(core0));
///       output('in core 0');
///       FOR(core1 << 0, core1 < 5, () => inc(core1));
///         output('   in core 1');
///       NEXT(); 
///     NEXT();
///     
/// # Related
/// * [NEXT]
/// * [WHILE]
/// * [ENDWHILE]
/// 
/// @param initializer initializes the memory location to the value that begins
/// the for/next sequence.
/// @param whileCondition is tested at the beginning of each iteration through the
/// FOR/NEXT loop, if the test returns false then the loop terminates.
/// @param incrementorFunc is a function that calls a given expression at the end
/// of the loop, just before returning to the top again.
/// 
void FOR(initializer, whileCondition, incrementorFunc()){
  whileCondition = _wrapLiteral(whileCondition);
  
  if (initializer != null){
    throw 'FOR initializer should be an assignment (ex. core0 << 0)';
  }
  
  _forStack.add(incrementorFunc);
  
  _globalContext.add('_for_context_${_nextLabelID}_');
  WHILE(whileCondition);
}


/// Declares the end of a FOR/NEXT loop block.
/// 
/// # Example
/// See the example in [FOR] to see how NEXT is used.
/// 
/// # Related
/// * [FOR]
/// * [WHILE]
/// * [ENDWHILE]
void NEXT(){
  if (_forStack.isEmpty){
    throw 'Found NEXT() without a corresponding FOR(). Check to make sure your FOR/NEXT declarations are balanced.';
  }
  
  if (_globalContext.isEmpty){
    throw 'Bad compiler state.  This is probably a bug on our end.';
  }
  
  
  final gctx = _globalContext.removeLast();
  final incrementor = _forStack.removeLast();
  incrementor();
  ENDWHILE();
  _MMU.dalloc('expression${gctx}');
  if (_globalContext.isEmpty){
    _MMU.dalloc('expression');
  }
}

/// Declares a condition loop that continues while the given [expr] evaluates to
/// true.
/// 
/// All WHILE() statements just eventually end with a [ENDWHILE] statement.
/// 
/// ## Limitations
/// Complex looping scenarios are not well supported.  Use care when nesting WHILE
/// loops, as this will likely yield unexpected results.  You may need to use the
/// advanced API if you require more complex looping implementations, or [inline]
/// your own hakkML directly.
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///          
///     // loop that ends when core0 >= 10
///     WHILE(core0 < 10);
///       output(core0);
///       inc(core0);
///     ENDWHILE();
///     
///     // infinite loop.  Very common in a hakk to have a main loop that
///     // is continuous.
///     WHILE(true);
///       output('to infinity and beyond!');
///     ENDWHILE();
/// 
/// # Related
/// * [ENDWHILE]
/// * [FOR]
/// * [NEXT]
void WHILE(expr){
  expr = _wrapLiteral(expr);
  
  _whileStack.add(_nextLabelID);
  _HakkObject.addLine('lbl: __while_test_${_whileStack.last}__');
  _globalContext.add('_while_context_${_nextLabelID}_');
  IF(expr);
}

/// Defines the end of a [WHILE] block.  When the [WHILE] expression evaluates to
/// false, code immediately following ENDWHILE will execute.
/// 
/// # Related
/// * [WHILE]
/// * [FOR]
/// * [NEXT]
void ENDWHILE(){
  if (_whileStack.isEmpty){
    throw 'Found ENDWHILE() without a corresponding WHILE(). Check to make sure your WHILE/ENDWHILE declarations are balanced.';
  }
  _HakkObject.addLine('goto: __while_test_${_whileStack.last}__');
  ENDIF();
  final gctx = _globalContext.removeLast();
  _MMU.dalloc('expression${gctx}');
  if (_globalContext.isEmpty){
    _MMU.dalloc('expression');
  }
  _whileStack.removeLast();
}

/// Provides a conditional mechanism that executes code inside the IF/ENDIF block 
/// when the given [expr] evaluates to true.
/// 
/// Any IF statement must have a corresponding [ENDIF].
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///     
///     core0 << 10
///     IF(core0 < 100);
///       output('core0 is less than 100');
///     ENDIF();
///     
/// # Related
/// * [ENDIF]
/// 
/// @param expr is the conditional expression to test.
void IF(expr){
  expr = _wrapLiteral(expr);
  _ifStack.add(_nextLabelID);
 
  final i = _ifStack.length;
  final li = _nextLabelID;

  final m = _MMU.malloc('if_context_${_ifStack.last}');
  
  if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('je: [${expr.memLoc}, true, __iftrue_${_ifStack.last}__]');
  }else{
    _HakkObject.addLine('set: [${m}, ${expr}]');
    _HakkObject.addLine('je: [${m}, true, __iftrue_${_ifStack.last}__]');    
  }
  _HakkObject.addLine('goto: __endif_${_ifStack.last}__');
  _HakkObject.addLine('lbl: __iftrue_${_ifStack.last}__');
  
  _globalContext.add('_if_context_${_nextLabelID}_');
}

/// Defines the end of a [IF] block.
/// 
/// # Related
/// * [IF]
void ENDIF(){
  if (_ifStack.isEmpty){
    throw 'Found ENDIF() without a corresponding IF(). Check to make sure your IF/ENDIF declarations are balanced.';
  }
  _HakkObject.addLine('lbl: __endif_${_ifStack.last}__');
  final ifctx = _ifStack.removeLast();
  _MMU.dalloc('if_context_${ifctx}');
  final gctx = _globalContext.removeLast();
  _MMU.dalloc('expression${gctx}');
  _MMU.dalloc('call_expression${gctx}');
  if (_globalContext.isEmpty){
    _MMU.dalloc('expression');
  }
}

/// Jumps code execution to the given [label].  A label can be a local [label] in
/// the current instruction block context, or it can be the name of another 
/// [Block.instruction], allowing transfer of control between instruction blocks.
/// 
/// ## Example
///     hakk('inline_example');
///     Block.system(0, 1);
///     
///     GOTO('this_place');
///     output('This will never be seen.');
///     
///     label('this_place');
///     output('Hi there.');
///     
/// @param label is the name of the label or instruction block to jump to.
void GOTO(String label){
  _HakkObject.addLine('goto: ${label}');
}

/// Initiates a call to a [SUB] subroutine, which is eventually [RETURN]ed to the
/// instruction immediately following the GOSUB statement.
/// 
/// # parameter passing and value returns
/// The GOSUB/SUB/RETURN sequence also supports a rudimentary parameter passing
/// and value return.
/// 
/// ## Subroutines are NOT closures.
/// All values stored in memory blocks (such as [core0] or any memory block added
/// to a hakk) are essential "global".  This means that you must plan carefully
/// the use of these memory locations to prevent "clobbering", where one operation
/// overwrites a memory location that was not intended.
/// 
/// ## param/argument matching
/// It is important that when using GOSUB the number of parameter values passed, must
/// match the number of arguements expected to be received by the [SUB]routine. 
///  
/// # GOSUB Must be used in an expression.
/// [SUB]routines always return a value, even if one is not provided.  The default
/// return value is 0, if none is provided.  A GOSUB should always be assigned to some
/// memory location.  It will not execute correctly if called alone.
/// 
/// ## This works
///     core0 << GOSUB('foo');
/// 
/// ## This won't work
///     GOSUB('foo');
/// 
/// ## If you want to call a [SUB] and do care about the return result, use [CALL].
///     CALL('foo');
///  
/// # Limitations
/// While GOSUB returns are well tolerated in expressions, they cannot at this time
/// be used directly as parameters.
/// 
/// ## Don't do this
///     GOSUB('foo', [GOSUB('bar'), GOSUB('bar')]);
///     
/// ## Do this instead
///     core0 << GOSUB('bar');
///     GOSUB('foo', [core0, core0]);
/// 
/// # Example
///     hakk('GOSUB_example');
///     Block.system(0, 1);
///     
///     // GOSUB to a subroutine with no parameters
///     core0 << GOSUB('say_hi');
///     
///     // Pass 2 parameters to 'adder' and store the
///     // returned result in core0.
///     core0 << GOSUB('adder', [5, 5]);
///     // "10"
///     output(core0);
///     
///     SUB('say_hi');
///       output('hi');
///     RETURN();
///     
///     SUB('adder', [core3, core4]);
///       core5 << core3 + core4;
///     RETURN(core5);
///     
///     output('Hi there.');
///     
/// # Related
/// * [SUB]
/// * [RETURN]
/// * [CALL]
/// 
/// @param subroutineName is the name of the [SUB]routine to call.
/// @param params optionally passes parameters to the [SUB]routine.
_ExpressionNode GOSUB(String subroutineName, [List params]){
  final node = new _ExpressionNode.call(subroutineName, params);
  _expressionQueue.add(node);
  return node;
}

/// A version of [GOSUB] that can be used outside of an expression.
/// 
/// # Example
///     hakk('CALL_example');
///     Block.system(0, 1);
///     
///     // CALL to a subroutine, but don't worry about the result returned.
///     CALL('foo');
///     
///     // Using GOSUB, we would have to assign it to a memory location or use it
///     // in an expression.
///     core0 << GOSUB('foo');
///     
///     // core0 now equals 0, because that is the default return value none is 
///     // provided by the called subroutine.
///     
///     output('finished');
///     
///     SUB('foo');
///       output('hello world!');
///     RETURN();
/// 
/// # Related
/// * [SUB]
/// * [RETURN]
/// * [GOSUB]
/// 
/// @param subroutineName is the name of the [SUB]routine to call.
/// @param params optionally passes parameters to the [SUB]routine.
void CALL(String subroutineName, [List params]){
  final node = new _ExpressionNode.call(subroutineName, params);
  _expressionQueue.add(node);
  node.compile();
}

/// Declares a subroutine block that can take 0-n arguments, run some code,
/// and then return a result back to the [GOSUB] that called it.
/// 
/// # Limitations
/// A Subroutine is NOT a closure.  All values and references used within it are
/// global to the hakk runtime context.
/// 
/// Subroutines do not support recursion (calling itself) or nesting (a subroutine 
/// within another subroutine).  Attempts to use them in these ways will likely 
/// yield unexpected results.  
/// 
/// It is possible to emulate these behaviors yourself, by using the advanced API, or
/// by [inline]ing hakkML.  Proceed with caution.
/// 
/// # Example
/// See the [GOSUB] documentation for examples of how SUB is used.
/// 
/// # Related
/// * [GOSUB]
/// * [RETURN]
/// * [CALL]
/// 
/// @param subroutineName is the declared name of the subroutine that [GOSUB] will use
/// to call it.
/// @param arguments is a list of memory locations that will store the values passed
/// by the [GOSUB] call.  Make sure the number of arguments matches the number of
/// parameters!
void SUB(String subroutineName, [List<_HakkObject> arguments]){
  if (_subroutineContext != null){
    throw 'Nested subroutines are not allowed.';
  }
  _subroutineContext = '${subroutineName}_protect';
  GOTO(_subroutineContext);
  label(subroutineName);
  if (arguments != null){
    arguments.reversed.forEach((a){
      if (a is! _HakkObject || (a is _HakkObject && a.isLiteral)){
        throw 'Argument must be a memory reference.';
      }
      a << pop();
    });
  }
}

/// Returns control from a [SUB]routine back to the [GOSUB] that called it.
/// 
/// RETURN may pass an optiona expression back to the GOSUB.
///
/// # Example
/// See the [GOSUB] documentation for examples of how RETURN is used.
///
/// # Related
/// * [SUB]
/// * [GOSUB]
void RETURN([expr]){
  if (_subroutineContext == null){
    throw 'Cannot have a subroutine [ret]urn without first declaring a subroutine().';
  }
  if (expr == null){
    _HakkObject.addLine('push: 0');
  }else{
    expr = _wrapLiteral(expr);
    push(expr);
  }
  
  _HakkObject.addLine('ret:');
  label(_subroutineContext);
  _subroutineContext = null;
}


/// Defines a label in code that can be used by [GOTO] to jump to a different
/// location in code.
/// 
/// # Example
/// See the [GOTO] documenation for examples of how label can be used.
/// 
/// # Related
/// * [GOTO]
/// 
/// @param label is the name which will be used by a [GOTO] statement to jump to this
/// location in the hakk.
void label(String label){
  _HakkObject.addLine('lbl: ${label}');
}


/// Unconditionally terminates processing of hakk code in the hakk.it virtual
/// machine.
/// 
/// This instruction is generally used in testing only.  It probably wouldn't help
/// your chances to win in the arenas if your hakk stops running!
/// 
/// @param expr is an expression that should result in a termination code for the
/// hakk.  At this time, this information is not exposed to the player, and can only
/// be inspected by internal tooling.
void EXIT(expr){
  expr = _wrapLiteral(expr);
  if (expr is _ExpressionNode){
    expr.compile();
    _HakkObject.addLine('exit: ${expr.memLoc}');
  }else{
    _HakkObject.addLine('exit: ${expr}');    
  }
}
